"""Defines a jax-compatible SimulationData."""
from __future__ import annotations

from typing import Tuple, Dict, Union, List

import pydantic.v1 as pd

from jax.tree_util import register_pytree_node_class

from .....components.data.monitor_data import MonitorDataType, FieldData, PermittivityData
from .....components.data.sim_data import SimulationData

from ..base import JaxObject
from ..simulation import JaxSimulation, JaxInfo
from .monitor_data import JaxMonitorDataType, JAX_MONITOR_DATA_MAP


@register_pytree_node_class
class JaxSimulationData(SimulationData, JaxObject):
    """A :class:`.SimulationData` registered with jax."""

    output_data: Tuple[JaxMonitorDataType, ...] = pd.Field(
        (),
        title="Jax Data",
        description="Tuple of Jax-compatible data associated with output monitors.",
        jax_field=True,
    )

    grad_data: Tuple[FieldData, ...] = pd.Field(
        (),
        title="Gradient Field Data",
        description="Tuple of monitor data storing fields associated with the input structures.",
    )

    grad_eps_data: Tuple[PermittivityData, ...] = pd.Field(
        (),
        title="Gradient Permittivity Data",
        description="Tuple of monitor data storing epsilon associated with the input structures.",
    )

    simulation: JaxSimulation = pd.Field(
        ...,
        title="Simulation",
        description="The jax-compatible simulation corresponding to the data.",
    )

    task_id: str = pd.Field(
        None,
        title="Task ID",
        description="Optional field storing the task_id for the original JaxSimulation.",
    )

    @property
    def grad_data_symmetry(self) -> Tuple[FieldData, ...]:
        """``self.grad_data`` but with ``symmetry_expanded_copy`` applied."""
        return tuple(data.symmetry_expanded_copy for data in self.grad_data)

    @property
    def grad_eps_data_symmetry(self) -> Tuple[FieldData, ...]:
        """``self.grad_eps_data`` but with ``symmetry_expanded_copy`` applied."""
        return tuple(data.symmetry_expanded_copy for data in self.grad_eps_data)

    @property
    def output_monitor_data(self) -> Dict[str, JaxMonitorDataType]:
        """Dictionary of ``.output_data`` monitor ``.name`` to the corresponding data."""
        return {monitor_data.monitor.name: monitor_data for monitor_data in self.output_data}

    @property
    def monitor_data(self) -> Dict[str, Union[JaxMonitorDataType, MonitorDataType]]:
        """Dictionary of ``.output_data`` monitor ``.name`` to the corresponding data."""
        reg_mnt_data = {monitor_data.monitor.name: monitor_data for monitor_data in self.data}
        reg_mnt_data.update(self.output_monitor_data)
        return reg_mnt_data

    @staticmethod
    def split_data(
        mnt_data: List[MonitorDataType], jax_info: JaxInfo
    ) -> Dict[str, List[MonitorDataType]]:
        """Split list of monitor data into data, output_data, grad_data, and grad_eps_data."""
        # Get information needed to split the full data list
        len_output_data = jax_info.num_output_monitors
        len_grad_data = jax_info.num_grad_monitors
        len_grad_eps_data = jax_info.num_grad_eps_monitors
        len_data = len(mnt_data) - len_output_data - len_grad_data - len_grad_eps_data

        # split the data list into regular data, output_data, and grad_data
        all_data = list(mnt_data)
        data = all_data[:len_data]
        output_data = all_data[len_data : len_data + len_output_data]
        grad_data = all_data[
            len_data + len_output_data : len_data + len_output_data + len_grad_data
        ]
        grad_eps_data = all_data[len_data + len_output_data + len_grad_data :]

        return dict(
            data=data, output_data=output_data, grad_data=grad_data, grad_eps_data=grad_eps_data
        )

    @classmethod
    def from_sim_data(
        cls, sim_data: SimulationData, jax_info: JaxInfo, task_id: str = None
    ) -> JaxSimulationData:
        """Construct a :class:`.JaxSimulationData` instance from a :class:`.SimulationData`."""

        self_dict = sim_data.dict(exclude={"type", "simulation", "data"})

        # convert the simulation to JaxSimulation
        jax_sim = JaxSimulation.from_simulation(simulation=sim_data.simulation, jax_info=jax_info)

        # construct JaxSimulationData with no data (yet)
        self_dict["simulation"] = jax_sim
        self_dict["data"] = ()

        data_dict = cls.split_data(mnt_data=sim_data.data, jax_info=jax_info)

        # convert the output data to the proper jax type
        output_data_list = []
        for mnt_data in data_dict["output_data"]:
            mnt_data_type_str = type(mnt_data)
            if mnt_data_type_str not in JAX_MONITOR_DATA_MAP:
                raise KeyError(
                    f"MonitorData type '{mnt_data_type_str}' "
                    "not currently supported by adjoint plugin."
                )
            mnt_data_type = JAX_MONITOR_DATA_MAP[mnt_data_type_str]
            jax_mnt_data = mnt_data_type.from_monitor_data(mnt_data)
            output_data_list.append(jax_mnt_data)
            data_dict["output_data"] = output_data_list

        self_dict.update(data_dict)
        self_dict.update(dict(task_id=task_id))

        return cls.parse_obj(self_dict)

    @classmethod
    def split_fwd_sim_data(
        cls, sim_data: SimulationData, jax_info: JaxInfo
    ) -> Tuple[SimulationData, SimulationData]:
        """Split a :class:`.SimulationData` into two parts, containing user and gradient data."""

        sim = sim_data.simulation

        data_dict = cls.split_data(mnt_data=sim_data.data, jax_info=jax_info)
        user_data = data_dict["data"] + data_dict["output_data"]
        adjoint_data = data_dict["grad_data"] + data_dict["grad_eps_data"]

        mnt_dict = JaxSimulation.split_monitors(
            monitors=sim_data.simulation.monitors, jax_info=jax_info
        )
        user_mnts = mnt_dict["monitors"] + mnt_dict["output_monitors"]
        adjoint_mnts = mnt_dict["grad_monitors"] + mnt_dict["grad_eps_monitors"]

        user_sim = sim.updated_copy(monitors=user_mnts)
        adjoint_sim = sim.updated_copy(monitors=adjoint_mnts)

        user_sim_data = sim_data.updated_copy(data=user_data, simulation=user_sim)
        adjoint_sim_data = sim_data.updated_copy(data=adjoint_data, simulation=adjoint_sim)

        return user_sim_data, adjoint_sim_data

    def make_adjoint_simulation(self, fwidth: float) -> JaxSimulation:
        """Make an adjoint simulation out of the data provided (generally, the vjp sim data)."""

        # grab boundary conditions with flipped bloch vectors (for adjoint)
        bc_adj = self.simulation.boundary_spec.flipped_bloch_vecs

        # add all adjoint sources and boundary conditions (at same time for BC validators to work)
        adj_srcs = []
        for mnt_data_vjp in self.output_data:
            for adj_source in mnt_data_vjp.to_adjoint_sources(fwidth=fwidth):
                adj_srcs.append(adj_source)

        update_dict = dict(boundary_spec=bc_adj, sources=adj_srcs, monitors=(), output_monitors=())
        update_dict.update(
            self.simulation.get_grad_monitors(
                input_structures=self.simulation.input_structures,
                freq_adjoint=self.simulation.freq_adjoint,
                include_eps_mnts=False,
            )
        )
        return self.simulation.updated_copy(**update_dict)



"""Tests Geometry objects."""

import pytest
import pydantic.v1 as pydantic
import numpy as np
import shapely
import matplotlib.pyplot as plt
import gdstk
import gdspy
import trimesh
import warnings

import tidy3d as td
from tidy3d.exceptions import SetupError, Tidy3dKeyError, ValidationError
from tidy3d.components.geometry.base import Planar
from tidy3d.components.geometry.utils import flatten_groups, traverse_geometries


GEO = td.Box(size=(1, 1, 1))
GEO_INF = td.Box(size=(1, 1, td.inf))
BOX = td.Box(size=(1, 1, 1))
BOX_2D = td.Box(size=(1, 0, 1))
POLYSLAB = td.PolySlab(vertices=((0, 0), (1, 0), (1, 1), (0, 1)), slab_bounds=(-0.5, 0.5), axis=2)
SPHERE = td.Sphere(radius=1)
CYLINDER = td.Cylinder(axis=2, length=1, radius=1)

GROUP = td.GeometryGroup(
    geometries=[
        td.Box(center=(-0.25, 0, 0), size=(0.5, 1, 1)),
        td.Box(center=(0.25, 0, 0), size=(0.5, 1, 1)),
    ]
)
UNION = td.ClipOperation(
    operation="union",
    geometry_a=td.Box(center=(-0.25, 0, 0), size=(0.5, 1, 1)),
    geometry_b=td.Box(center=(0.25, 0, 0), size=(0.5, 1, 1)),
)
INTERSECTION = td.ClipOperation(operation="intersection", geometry_a=UNION, geometry_b=SPHERE)
DIFFERENCE = td.ClipOperation(operation="difference", geometry_a=CYLINDER, geometry_b=BOX)
SYM_DIFFERENCE = td.ClipOperation(
    operation="symmetric_difference",
    geometry_a=td.ClipOperation(
        operation="difference",
        geometry_a=td.Box(size=(td.inf, td.inf, td.inf)),
        geometry_b=td.Box(center=(-0.25, 0, 0), size=(0.5, 1, 1)),
    ),
    geometry_b=td.ClipOperation(
        operation="difference",
        geometry_a=td.Box(size=(td.inf, td.inf, td.inf)),
        geometry_b=td.Box(center=(0.25, 0, 0), size=(0.5, 1, 1)),
    ),
)


GEO_TYPES = [
    BOX,
    CYLINDER,
    SPHERE,
    POLYSLAB,
    UNION,
    INTERSECTION,
    DIFFERENCE,
    SYM_DIFFERENCE,
    GROUP,
]

_, AX = plt.subplots()


@pytest.mark.parametrize("component", GEO_TYPES)
def test_plot(component):
    _ = component.plot(z=0, ax=AX)
    plt.close()


def test_base_inside():
    assert td.Geometry.inside(GEO, x=0, y=0, z=0)
    assert np.all(td.Geometry.inside(GEO, np.array([0, 0]), np.array([0, 0]), np.array([0, 0])))
    assert np.all(
        td.Geometry.inside(GEO, np.array([[0, 0]]), np.array([[0, 0]]), np.array([[0, 0]]))
    )


def test_base_inside_meshgrid():
    assert np.all(td.Geometry.inside_meshgrid(GEO, x=[0], y=[0], z=[0]))
    assert np.all(td.Geometry.inside_meshgrid(GEO, [0, 0], [0, 0], [0, 0]))
    # Input dimensions different than 1 error for ``inside_meshgrid``.
    with pytest.raises(ValueError):
        _ = td.Geometry.inside_meshgrid(GEO, x=0, y=0, z=0)
    with pytest.raises(ValueError):
        _ = td.Geometry.inside_meshgrid(GEO, [[0, 0]], [[0, 0]], [[0, 0]])


def test_bounding_box():
    assert GEO.bounding_box == GEO
    assert GEO_INF.bounding_box == GEO_INF


@pytest.mark.parametrize("points_shape", [(3,), (3, 10)])
def test_rotate_points(points_shape):
    points = np.random.random(points_shape)
    points_rotated = td.Geometry.rotate_points(points=points, axis=(0, 0, 1), angle=2 * np.pi)
    assert np.allclose(points, points_rotated)
    points_rotated = td.Geometry.rotate_points(points=points, axis=(0, 0, 1), angle=np.pi)


@pytest.mark.parametrize("axis", [0, 1, 2])
def test_reflect_points(axis):
    points = np.random.random((3, 10))
    pr = GEO.reflect_points(points=points, polar_axis=2, angle_theta=2 * np.pi, angle_phi=0)
    assert np.allclose(pr, points)
    pr = GEO.reflect_points(points=points, polar_axis=2, angle_theta=0, angle_phi=2 * np.pi)
    assert np.allclose(pr, points)


@pytest.mark.parametrize("component", GEO_TYPES)
def test_volume(component):
    _ = component.volume()
    _ = component.volume(bounds=GEO.bounds)
    _ = component.volume(bounds=((-100, -100, -100), (100, 100, 100)))
    _ = component.volume(bounds=((-0.1, -0.1, -0.1), (0.1, 0.1, 0.1)))
    _ = component.volume(bounds=((-100, -100, -100), (-10, -10, -10)))
    _ = component.volume(bounds=((10, 10, 10), (100, 100, 100)))


@pytest.mark.parametrize("component", GEO_TYPES)
def test_surface_area(component):
    _ = component.surface_area()
    _ = component.surface_area(bounds=GEO.bounds)
    _ = component.surface_area(bounds=((-100, -100, -100), (100, 100, 100)))
    _ = component.surface_area(bounds=((-0.1, -0.1, -0.1), (0.1, 0.1, 0.1)))
    _ = component.surface_area(bounds=((-100, -100, -100), (-10, -10, -10)))
    _ = component.surface_area(bounds=((10, 10, 10), (100, 100, 100)))


@pytest.mark.parametrize("component", GEO_TYPES)
def test_bounds(component):
    _ = component.bounds


def test_planar_bounds():
    _ = Planar.bounds.fget(CYLINDER)


@pytest.mark.parametrize("component", GEO_TYPES)
def test_inside(component):
    _ = component.inside(0, 0, 0)
    _ = component.inside(np.array([0, 0]), np.array([0, 0]), np.array([0, 0]))
    _ = component.inside(np.array([[0, 0]]), np.array([[0, 0]]), np.array([[0, 0]]))


def test_zero_dims():
    assert BOX.zero_dims == []
    assert BOX_2D.zero_dims == [1]


def test_inside_polyslab_sidewall():
    ps = POLYSLAB.copy(update=dict(sidewall_angle=0.1))
    ps.inside(x=0, y=0, z=0)


# TODO: Weiliang fix this test? doesnt work when sidewall non-zero
def test_inside_polyslab_sidewall_arrays():
    inside_kwargs = {coord: np.array([-1, 0, 1]) for coord in "xyz"}
    POLYSLAB.inside(**inside_kwargs)
    # ps = POLYSLAB.copy(update=dict(sidewall_angle=0.1))
    # ps.inside(**inside_kwargs)


def test_array_to_vertices():
    vertices = ((0, 0), (1, 0), (1, 1))
    array = POLYSLAB.vertices_to_array(vertices)
    vertices2 = POLYSLAB.array_to_vertices(array)
    assert np.all(np.array(vertices) == np.array(vertices2))


@pytest.mark.parametrize("component", GEO_TYPES)
def test_intersections_plane(component):
    assert len(component.intersections_plane(z=0.2)) > 0
    assert len(component.intersections_plane(x=0.2)) > 0
    assert len(component.intersections_plane(x=10000)) == 0


def test_bounds_base():
    assert all(a == b for a, b in zip(Planar.bounds.fget(POLYSLAB), POLYSLAB.bounds))


def test_center_not_inf_validate():
    with pytest.raises(pydantic.ValidationError):
        _ = td.Box(center=(td.inf, 0, 0))
    with pytest.raises(pydantic.ValidationError):
        _ = td.Box(center=(-td.inf, 0, 0))


def test_radius_not_inf_validate():
    with pytest.raises(pydantic.ValidationError):
        _ = td.Sphere(radius=td.inf)
    with pytest.raises(pydantic.ValidationError):
        _ = td.Cylinder(radius=td.inf, center=(0, 0, 0), axis=1, length=1)


def test_slanted_cylinder_infinite_length_validate():
    _ = td.Cylinder(radius=1, center=(0, 0, 0), axis=1, length=td.inf)
    _ = td.Cylinder(radius=1, center=(0, 0, 0), axis=1, length=td.inf, reference_plane="top")
    _ = td.Cylinder(radius=1, center=(0, 0, 0), axis=1, length=td.inf, reference_plane="bottom")
    _ = td.Cylinder(radius=1, center=(0, 0, 0), axis=1, length=td.inf, reference_plane="middle")
    _ = td.Cylinder(
        radius=1,
        center=(0, 0, 0),
        axis=1,
        length=td.inf,
        sidewall_angle=0.1,
        reference_plane="middle",
    )
    with pytest.raises(pydantic.ValidationError):
        _ = td.Cylinder(
            radius=1,
            center=(0, 0, 0),
            axis=1,
            length=td.inf,
            sidewall_angle=0.1,
            reference_plane="top",
        )
    with pytest.raises(pydantic.ValidationError):
        _ = td.Cylinder(
            radius=1,
            center=(0, 0, 0),
            axis=1,
            length=td.inf,
            sidewall_angle=0.1,
            reference_plane="bottom",
        )


def test_box_from_bounds():
    b = td.Box.from_bounds(rmin=(-td.inf, 0, 0), rmax=(td.inf, 0, 0))
    assert b.center[0] == 0.0

    with pytest.raises(SetupError):
        _ = td.Box.from_bounds(rmin=(0, 0, 0), rmax=(td.inf, 0, 0))

    b = td.Box.from_bounds(rmin=(-1, -1, -1), rmax=(1, 1, 1))
    assert b.center == (0, 0, 0)


def test_polyslab_center_axis():
    """Test the handling of center_axis in a polyslab having (-td.inf, td.inf) bounds."""
    ps = POLYSLAB.copy(update=dict(slab_bounds=(-td.inf, td.inf)))
    assert ps.center_axis == 0


@pytest.mark.parametrize(
    "lower_bound, upper_bound", ((-td.inf, td.inf), (-1, td.inf), (-td.inf, 1))
)
def test_polyslab_inf_bounds(lower_bound, upper_bound):
    """Test the handling of various operations in a polyslab having inf bounds."""
    ps = POLYSLAB.copy(update=dict(slab_bounds=(lower_bound, upper_bound)))
    # catch any runtime warning related to inf operations
    with warnings.catch_warnings():
        warnings.simplefilter("error")
        bounds = ps.bounds
        ps.intersections_plane(x=0.5)
        ps.intersections_plane(z=0)


def test_polyslab_bounds():
    with pytest.raises(pydantic.ValidationError):
        td.PolySlab(vertices=((0, 0), (1, 0), (1, 1)), slab_bounds=(0.5, -0.5), axis=2)


def test_validate_polyslab_vertices_valid():
    with pytest.raises(pydantic.ValidationError):
        POLYSLAB.copy(update=dict(vertices=(1, 2, 3)))
    with pytest.raises(pydantic.ValidationError):
        crossing_verts = ((0, 0), (1, 1), (0, 1), (1, 0))
        POLYSLAB.copy(update=dict(vertices=crossing_verts))


def test_sidewall_failed_validation():
    with pytest.raises(pydantic.ValidationError):
        POLYSLAB.copy(update=dict(sidewall_angle=1000))


def test_surfaces():
    with pytest.raises(SetupError):
        td.Box.surfaces(size=(1, 0, 1), center=(0, 0, 0))

    td.FluxMonitor.surfaces(
        size=(1, 1, 1), center=(0, 0, 0), normal_dir="+", name="test", freqs=[1]
    )
    td.Box.surfaces(size=(1, 1, 1), center=(0, 0, 0), normal_dir="+")


def test_arrow_both_dirs():
    _, ax = plt.subplots()
    GEO._plot_arrow(direction=(1, 2, 3), x=0, both_dirs=True, ax=ax)


def test_gdstk_cell():
    gds_cell = gdstk.Cell("name")
    gds_cell.add(gdstk.rectangle((0, 0), (1, 1)))
    td.PolySlab.from_gds(gds_cell=gds_cell, axis=2, slab_bounds=(-1, 1), gds_layer=0)
    td.PolySlab.from_gds(gds_cell=gds_cell, axis=2, slab_bounds=(-1, 1), gds_layer=0, gds_dtype=0)
    with pytest.raises(Tidy3dKeyError):
        td.PolySlab.from_gds(gds_cell=gds_cell, axis=2, slab_bounds=(-1, 1), gds_layer=1)
    with pytest.raises(Tidy3dKeyError):
        td.PolySlab.from_gds(
            gds_cell=gds_cell, axis=2, slab_bounds=(-1, 1), gds_layer=1, gds_dtype=0
        )


def test_gdspy_cell():
    gds_cell = gdspy.Cell("name")
    gds_cell.add(gdspy.Rectangle((0, 0), (1, 1)))
    td.PolySlab.from_gds(gds_cell=gds_cell, axis=2, slab_bounds=(-1, 1), gds_layer=0)
    with pytest.raises(Tidy3dKeyError):
        td.PolySlab.from_gds(gds_cell=gds_cell, axis=2, slab_bounds=(-1, 1), gds_layer=1)


def make_geo_group():
    """Make a generic Geometry Group."""
    boxes = [td.Box(size=(1, 1, 1), center=(i, 0, 0)) for i in range(-5, 5)]
    return td.GeometryGroup(geometries=boxes)


def test_geo_group_initialize():
    """make sure you can construct one."""
    _ = make_geo_group()


def test_geo_group_structure():
    """make sure you can construct a structure using GeometryGroup."""
    geo_group = make_geo_group()
    _ = td.Structure(geometry=geo_group, medium=td.Medium())


def test_geo_group_methods():
    """Tests the geometry methods of geo group."""
    geo_group = make_geo_group()
    geo_group.inside(0, 1, 2)
    geo_group.inside(np.linspace(0, 1, 10), np.linspace(0, 1, 10), np.linspace(0, 1, 10))
    geo_group.inside_meshgrid(np.linspace(0, 1, 10), np.linspace(0, 1, 10), np.linspace(0, 1, 10))
    geo_group.intersections_plane(y=0)
    geo_group.intersects(td.Box(size=(1, 1, 1)))
    _ = geo_group.bounds


def test_geo_group_empty():
    """dont allow empty geometry list."""
    with pytest.raises(pydantic.ValidationError):
        _ = td.GeometryGroup(geometries=[])


def test_geo_group_volume():
    geo_group = make_geo_group()
    geo_group.volume(bounds=GEO.bounds)


def test_geo_group_surface_area():
    geo_group = make_geo_group()
    geo_group.surface_area(bounds=GEO.bounds)


def test_geometryoperations():
    assert BOX + CYLINDER == td.GeometryGroup(geometries=(BOX, CYLINDER))
    assert BOX + UNION == td.GeometryGroup(geometries=(BOX, UNION.geometry_a, UNION.geometry_b))
    assert UNION + CYLINDER == td.GeometryGroup(
        geometries=(UNION.geometry_a, UNION.geometry_b, CYLINDER)
    )
    assert BOX + GROUP == td.GeometryGroup(geometries=(BOX,) + GROUP.geometries)
    assert GROUP + CYLINDER == td.GeometryGroup(geometries=GROUP.geometries + (CYLINDER,))

    assert BOX | CYLINDER == td.GeometryGroup(geometries=(BOX, CYLINDER))
    assert BOX | UNION == td.GeometryGroup(geometries=(BOX, UNION.geometry_a, UNION.geometry_b))
    assert UNION | CYLINDER == td.GeometryGroup(
        geometries=(UNION.geometry_a, UNION.geometry_b, CYLINDER)
    )
    assert BOX | GROUP == td.GeometryGroup(geometries=(BOX,) + GROUP.geometries)
    assert GROUP | CYLINDER == td.GeometryGroup(geometries=GROUP.geometries + (CYLINDER,))

    assert BOX * SPHERE == td.ClipOperation(
        operation="intersection", geometry_a=BOX, geometry_b=SPHERE
    )

    assert BOX & SPHERE == td.ClipOperation(
        operation="intersection", geometry_a=BOX, geometry_b=SPHERE
    )

    assert BOX - SPHERE == td.ClipOperation(
        operation="difference", geometry_a=BOX, geometry_b=SPHERE
    )

    assert BOX ^ SPHERE == td.ClipOperation(
        operation="symmetric_difference", geometry_a=BOX, geometry_b=SPHERE
    )


def test_flattening():
    flat = list(
        flatten_groups(
            td.GeometryGroup(
                geometries=[
                    td.Box(size=(1, 1, 1)),
                    td.Box(size=(0, 1, 0)),
                    td.ClipOperation(
                        operation="union",
                        geometry_a=td.Box(size=(0, 0, 1)),
                        geometry_b=td.GeometryGroup(
                            geometries=[
                                td.Box(size=(2, 2, 2)),
                                td.GeometryGroup(
                                    geometries=[td.Box(size=(3, 3, 3)), td.Box(size=(3, 0, 3))]
                                ),
                            ]
                        ),
                    ),
                ]
            )
        )
    )
    assert len(flat) == 6
    assert all(isinstance(g, td.Box) for g in flat)

    flat = list(
        flatten_groups(
            td.GeometryGroup(
                geometries=[
                    td.Box(size=(1, 1, 1)),
                    td.Box(size=(0, 1, 0)),
                    td.ClipOperation(
                        operation="intersection",
                        geometry_a=td.Box(size=(0, 0, 1)),
                        geometry_b=td.GeometryGroup(
                            geometries=[
                                td.Box(size=(2, 2, 2)),
                                td.GeometryGroup(
                                    geometries=[td.Box(size=(3, 3, 3)), td.Box(size=(3, 0, 3))]
                                ),
                            ]
                        ),
                    ),
                ]
            )
        )
    )
    assert len(flat) == 3
    assert all(
        isinstance(g, td.Box) or (isinstance(g, td.ClipOperation) and g.operation == "intersection")
        for g in flat
    )


def test_geometry_traversal():
    geometries = list(traverse_geometries(td.Box(size=(1, 1, 1))))
    assert len(geometries) == 1

    geo_tree = td.GeometryGroup(
        geometries=[
            td.Box(size=(1, 0, 0)),
            td.ClipOperation(
                operation="intersection",
                geometry_a=td.GeometryGroup(
                    geometries=[
                        td.Box(size=(5, 0, 0)),
                        td.Box(size=(6, 0, 0)),
                    ]
                ),
                geometry_b=td.ClipOperation(
                    operation="difference",
                    geometry_a=td.Box(size=(7, 0, 0)),
                    geometry_b=td.Box(size=(8, 0, 0)),
                ),
            ),
            td.GeometryGroup(
                geometries=[
                    td.Box(size=(3, 0, 0)),
                    td.Box(size=(4, 0, 0)),
                ]
            ),
            td.Box(size=(2, 0, 0)),
        ]
    )
    geometries = list(traverse_geometries(geo_tree))
    assert len(geometries) == 13


""" geometry """


def test_geometry():
    _ = td.Box(size=(1, 1, 1), center=(0, 0, 0))
    _ = td.Sphere(radius=1, center=(0, 0, 0))
    _ = td.Cylinder(radius=1, center=(0, 0, 0), axis=1, length=1)
    _ = td.PolySlab(vertices=((1, 2), (3, 4), (5, 4)), slab_bounds=(-1, 1), axis=2)
    # vertices_np = np.array(s.vertices)
    # _ = PolySlab(vertices=vertices_np, slab_bounds=(-1, 1), axis=1)

    # make sure wrong axis arguments error
    with pytest.raises(pydantic.ValidationError):
        _ = td.Cylinder(radius=1, center=(0, 0, 0), axis=-1, length=1)
    with pytest.raises(pydantic.ValidationError):
        _ = td.PolySlab(radius=1, center=(0, 0, 0), axis=-1, slab_bounds=(-0.5, 0.5))
    with pytest.raises(pydantic.ValidationError):
        _ = td.Cylinder(radius=1, center=(0, 0, 0), axis=3, length=1)
    with pytest.raises(pydantic.ValidationError):
        _ = td.PolySlab(radius=1, center=(0, 0, 0), axis=3, slab_bounds=(-0.5, 0.5))

    # make sure negative values error
    with pytest.raises(pydantic.ValidationError):
        _ = td.Sphere(radius=-1, center=(0, 0, 0))
    with pytest.raises(pydantic.ValidationError):
        _ = td.Cylinder(radius=-1, center=(0, 0, 0), axis=3, length=1)
    with pytest.raises(pydantic.ValidationError):
        _ = td.Cylinder(radius=1, center=(0, 0, 0), axis=3, length=-1)


def test_geometry_sizes():

    # negative in size kwargs errors
    for size in (-1, 1, 1), (1, -1, 1), (1, 1, -1):
        with pytest.raises(pydantic.ValidationError):
            _ = td.Box(size=size, center=(0, 0, 0))
        with pytest.raises(pydantic.ValidationError):
            _ = td.Simulation(size=size, run_time=1e-12, grid_spec=td.GridSpec(wavelength=1.0))

    # negative grid sizes error?
    with pytest.raises(pydantic.ValidationError):
        _ = td.Simulation(size=(1, 1, 1), grid_spec=td.GridSpec.uniform(dl=-1.0), run_time=1e-12)


@pytest.mark.parametrize("x0", [5])
def test_geometry_touching_intersections_plane(x0):
    """Two touching boxes should show at least one intersection at plane where they touch."""

    # size of each box
    # L = 1 # works
    # L = 0.1 # works
    # L = 0.12 # assertion errors
    L = 0.24  # assertion errors
    # L = 0.25 # works

    # one box to the left of x0 and one box to the right of x0, touching at x0
    b1 = td.Box(center=(x0 - L / 2, 0, 0), size=(L, L, L))
    b2 = td.Box(center=(x0 + L / 2, 0, 0), size=(L, L, L))

    ints1 = b1.intersections_plane(x=x0)
    ints2 = b2.intersections_plane(x=x0)

    ints_total = ints1 + ints2

    assert len(ints_total) > 0, "no intersections found at plane where two boxes touch"


def test_pop_axis():
    b = td.Box(size=(1, 1, 1))
    for axis in range(3):
        coords = (1, 2, 3)
        Lz, (Lx, Ly) = b.pop_axis(coords, axis=axis)
        _coords = b.unpop_axis(Lz, (Lx, Ly), axis=axis)
        assert all(c == _c for (c, _c) in zip(coords, _coords))
        _Lz, (_Lx, _Ly) = b.pop_axis(_coords, axis=axis)
        assert Lz == _Lz
        assert Lx == _Lx
        assert Ly == _Ly


def test_2b_box_intersections():
    plane = td.Box(size=(1, 4, 0))
    box1 = td.Box(size=(1, 1, 1))
    box2 = td.Box(size=(1, 1, 1), center=(3, 0, 0))

    result = plane.intersections_with(box1)
    assert len(result) == 1
    assert result[0].geom_type == "Polygon"
    assert len(plane.intersections_with(box2)) == 0

    with pytest.raises(ValidationError):
        _ = box1.intersections_with(box2)

    assert len(box1.intersections_2dbox(plane)) == 1
    assert len(box2.intersections_2dbox(plane)) == 0

    with pytest.raises(ValidationError):
        _ = box2.intersections_2dbox(box1)


def test_polyslab_merge():
    """make sure polyslabs from gds get merged when they should."""

    def make_polyslabs(gap_size):
        """Construct two rectangular polyslabs separated by a gap."""
        cell = gdstk.Cell(f"polygons_{gap_size:.2f}")
        rect1 = gdstk.rectangle((gap_size / 2, 0), (1, 1))
        rect2 = gdstk.rectangle((-1, 0), (-gap_size / 2, 1))
        cell.add(rect1, rect2)
        return td.PolySlab.from_gds(gds_cell=cell, gds_layer=0, axis=2, slab_bounds=(-1, 1))

    polyslabs_gap = make_polyslabs(gap_size=0.3)
    assert len(polyslabs_gap) == 2, "untouching polylsabs were merged incorrectly."

    polyslabs_touching = make_polyslabs(gap_size=0)
    assert len(polyslabs_touching) == 1, "polyslabs didnt merge correctly."


def test_polyslab_side_plot_merge():
    """In side plot, make sure splitted polygons merge."""
    x0 = 2
    y0 = 4
    z0 = 1
    R = 5
    wg_width = 0.5
    wg_thickness = 0.22
    sidewall_angle = 15 * np.pi / 180

    cell = gdstk.Cell("bottom")
    path_bottom = gdstk.RobustPath(
        (x0 + R, y0), wg_width - wg_thickness * np.tan(np.abs(sidewall_angle)), layer=1, datatype=0
    )

    path_bottom.arc(R, 0, -np.pi)
    cell.add(path_bottom)
    ring_bottom_geo = td.PolySlab.from_gds(
        cell,
        gds_layer=1,
        axis=2,
        slab_bounds=(z0 - wg_thickness / 2, z0 + wg_thickness / 2),
        sidewall_angle=sidewall_angle,
        reference_plane="top",
    )
    assert len(ring_bottom_geo[0].intersections_plane(x=2)) == 1


@pytest.mark.parametrize("axis", [0, 1, 2])
def test_polyslab_axis(axis):
    ps = td.PolySlab(slab_bounds=(-1, 1), vertices=((-5, -5), (-5, 5), (5, 5), (5, -5)), axis=axis)

    # bound test
    bounds_ideal = [-5, -5]
    bounds_ideal.insert(axis, -1)
    bounds_ideal = np.array(bounds_ideal)
    np.allclose(ps.bounds[0], bounds_ideal)
    np.allclose(ps.bounds[1], -bounds_ideal)

    # inside
    point = [0, 0]
    point.insert(axis, 3)
    assert not ps.inside(point[0], point[1], point[2])

    # intersections
    plane_coord = [None] * 3
    plane_coord[axis] = 3
    assert not ps.intersects_plane(x=plane_coord[0], y=plane_coord[1], z=plane_coord[2])
    plane_coord[axis] = -3
    assert not ps.intersects_plane(x=plane_coord[0], y=plane_coord[1], z=plane_coord[2])


def test_from_shapely():
    ring = shapely.LinearRing([(-16, 9), (-8, 9), (-12, 2)])
    poly = shapely.Polygon([(-2, 0), (-10, 0), (-6, 7)])
    hole = shapely.Polygon(
        [(0, 0), (9, 0), (9, 9), (0, 9), (0, 2), (2, 2), (2, 7), (7, 7), (7, 2), (0, 2)]
    ).buffer(0)
    collection = shapely.GeometryCollection((shapely.MultiPolygon((poly,)), hole, ring))

    geo = td.Geometry.from_shapely(collection, 2, (0, 1))
    assert len(geo.intersections_plane(z=0.5)) == 3

    geo = td.Geometry.from_shapely(
        collection, 2, (0, 1), sidewall_angle=1.0, reference_plane="bottom"
    )
    assert len(geo.intersections_plane(z=0)) == 3
    assert len(geo.intersections_plane(z=1)) == 2

    geo = td.Geometry.from_shapely(
        collection, 2, (0, 1), sidewall_angle=-1.0, reference_plane="top"
    )
    assert len(geo.intersections_plane(z=0)) == 2
    assert len(geo.intersections_plane(z=1)) == 3


def test_from_gds():
    ring = gdstk.Polygon([(-16, 9), (-8, 9), (-12, 2)], layer=1)
    poly = gdstk.Polygon([(-2, 0), (-10, 0), (-6, 7)])
    hole = gdstk.Polygon(
        [(0, 0), (9, 0), (9, 9), (0, 9), (0, 2), (2, 2), (2, 7), (7, 7), (7, 2), (0, 2)]
    )
    cell = gdstk.Cell("CELL").add(ring, poly, hole)
    geo = td.Geometry.from_gds(
        cell, 2, (0, 1), gds_layer=0, dilation=-0.5, sidewall_angle=0.5, reference_plane="bottom"
    )
    assert len(geo.intersections_plane(z=0)) == 2
    assert len(geo.intersections_plane(z=1)) == 1


def test_custom_surface_geometry(tmp_path):
    # create tetrahedron STL
    vertices = np.array([[0, 0, 0], [1, 0, 0], [0, 1, 0], [0, 0, 1]])
    faces = np.array([[1, 2, 3], [0, 3, 2], [0, 1, 3], [0, 2, 1]])
    tetrahedron = trimesh.Trimesh(vertices, faces)
    geom = td.TriangleMesh.from_trimesh(tetrahedron)

    # test import
    import_geom = td.TriangleMesh.from_stl("tests/data/tetrahedron.stl")
    assert np.allclose(import_geom.triangles, geom.triangles)

    # test export and then import
    geom.trimesh.export(str(tmp_path / "export.stl"))
    import_geom = td.TriangleMesh.from_stl(str(tmp_path / "export.stl"))
    assert np.allclose(import_geom.triangles, geom.triangles)

    # assert np.array_equal(tetrahedron.vectors, export_vectors)

    areas = [0.5 * np.sqrt(2) * np.sqrt(1 + 2 * 0.5**2), 0.5, 0.5, 0.5]
    unit_normals_unnormalized = [[1, 1, 1], [-1, 0, 0], [0, -1, 0], [0, 0, -1]]
    unit_normals = [n / np.linalg.norm(n) for n in unit_normals_unnormalized]
    _ = [n * a for (n, a) in zip(unit_normals, areas)]

    # test bounds
    assert np.allclose(np.array(geom.bounds), [[0, 0, 0], [1, 1, 1]])

    # test surface area
    assert np.isclose(geom.surface_area(), np.sum(areas))

    # test volume
    assert np.isclose(geom.volume(), 1 / 6)

    # test intersections
    assert shapely.equals(geom.intersections_plane(x=0), shapely.Polygon([[0, 0], [0, 1], [1, 0]]))
    assert shapely.equals(
        geom.intersections_plane(z=0.5), shapely.Polygon([[0, 0], [0, 0.5], [0.5, 0]])
    )

    # test inside
    assert geom.inside([0.2], [0.2], [0.2])[0]
    assert not geom.inside([0.8], [0.2], [0.2])[0]

    # test plot
    _, ax = plt.subplots()
    _ = geom.plot(z=0.1, ax=ax)
    plt.close()

    # test inconsistent winding
    vertices = np.array([[0, 0, 0], [1, 0, 0], [0, 1, 0], [0, 0, 1]])
    faces = np.array([[2, 1, 3], [0, 3, 2], [0, 1, 3], [0, 2, 1]])
    tetrahedron = trimesh.Trimesh(vertices, faces)
    # we currently just log a warning
    # with pytest.raises(ValidationError):
    geom = td.TriangleMesh.from_trimesh(tetrahedron)

    # test non-watertight mesh
    vertices = np.array([[0, 0, 0], [1, 0, 0], [0, 1, 0], [0, 0, 1]])
    faces = np.array([[0, 3, 2], [0, 1, 3], [0, 2, 1]])
    tetrahedron = trimesh.Trimesh(vertices, faces)
    # we currently just log a warning
    # with pytest.raises(ValidationError):
    geom = td.TriangleMesh.from_trimesh(tetrahedron)

    # test zero area triangles
    vertices = np.array([[1, 0, 0], [1, 0, 0], [0, 1, 0], [0, 0, 1]])
    faces = np.array([[1, 2, 3], [0, 3, 2], [0, 1, 3], [0, 2, 1]])
    tetrahedron = trimesh.Trimesh(vertices, faces)
    with pytest.raises(pydantic.ValidationError):
        geom = td.TriangleMesh.from_trimesh(tetrahedron)

    # test trimesh.Scene
    import_geom = td.TriangleMesh.from_stl("tests/data/two_boxes_separate.stl")
    sim = sim = td.Simulation(
        size=(10, 10, 10),
        grid_spec=td.GridSpec.uniform(dl=0.1),
        sources=[],
        structures=[td.Structure(geometry=import_geom, medium=td.Medium(permittivity=2))],
        monitors=[],
        run_time=1e-12,
        boundary_spec=td.BoundarySpec.all_sides(td.PML()),
    )
    _, ax = plt.subplots()
    _ = sim.plot(y=0, ax=ax)
    plt.close()


def test_geo_group_sim():

    geo_grp = td.TriangleMesh.from_stl("tests/data/two_boxes_separate.stl")
    geos_orig = list(geo_grp.geometries)
    geo_grp_full = geo_grp.updated_copy(geometries=geos_orig + [td.Box(size=(1, 1, 1))])

    sim = td.Simulation(
        size=(10, 10, 10),
        grid_spec=td.GridSpec.uniform(dl=0.1),
        sources=[],
        structures=[td.Structure(geometry=geo_grp_full, medium=td.Medium(permittivity=2))],
        monitors=[],
        run_time=1e-12,
        boundary_spec=td.BoundarySpec.all_sides(td.PML()),
    )

    # why is this failing?  assert 4==2
    assert len(sim.custom_datasets) == len(geos_orig)



"""Migrate authentication to API key."""
import json
import os

import click
import requests
import toml

from .constants import CONFIG_FILE, CREDENTIAL_FILE, TIDY3D_DIR
from ..environment import Env


def migrate() -> bool:
    """Click command to migrate the credential to api key."""
    if os.path.exists(CREDENTIAL_FILE):
        with open(CREDENTIAL_FILE, encoding="utf-8") as fp:
            auth_json = json.load(fp)
        email = auth_json["email"]
        password = auth_json["password"]
        if email and password:
            is_migrate = click.prompt(
                "This system was found to use the old authentication protocol based on auth.json, "
                "which will not be supported in the upcoming 2.0 release. We strongly recommend "
                "migrating to the API key authentication before the release. Would you like to "
                "migrate to the API key authentication now? "
                "This will create a '~/.tidy3d/config' file on your machine "
                "to store the API key from your online account but all other "
                "workings of Tidy3D will remain the same.",
                type=bool,
                default=True,
            )
            if is_migrate:
                headers = {"Application": "TIDY3D"}
                resp = requests.get(
                    f"{Env.current.web_api_endpoint}/auth",
                    headers=headers,
                    auth=(email, password),
                )
                if resp.status_code != 200:
                    click.echo(f"Migrate to api key failed: {resp.text}")
                    return False
                else:
                    # click.echo(json.dumps(resp.json(), indent=4))
                    access_token = resp.json()["data"]["auth"]["accessToken"]
                    headers["Authorization"] = f"Bearer {access_token}"
                    resp = requests.get(f"{Env.current.web_api_endpoint}/apikey", headers=headers)
                    if resp.status_code != 200:
                        click.echo(f"Migrate to api key failed: {resp.text}")
                        return False
                    else:
                        click.echo(json.dumps(resp.json(), indent=4))
                        apikey = resp.json()["data"]
                        if not apikey:
                            resp = requests.post(
                                f"{Env.current.web_api_endpoint}/apikey", headers=headers
                            )
                            if resp.status_code != 200:
                                click.echo(f"Migrate to api key failed: {resp.text}")
                                return False
                            else:
                                apikey = resp.json()["data"]
                        if not os.path.exists(TIDY3D_DIR):
                            os.mkdir(TIDY3D_DIR)
                        with open(CONFIG_FILE, "w+", encoding="utf-8") as config_file:
                            toml_config = toml.loads(config_file.read())
                            toml_config.update({"apikey": apikey})
                            config_file.write(toml.dumps(toml_config))

                        # rename auth.json to auth.json.bak
                        os.rename(CREDENTIAL_FILE, CREDENTIAL_FILE + ".bak")
                        return True
            else:
                click.echo("You can migrate to api key by running 'tidy3d migrate' command.")



"""Environment Setup."""
import os

from pydantic.v1 import BaseSettings, Field

from tidy3d import log


class EnvironmentConfig(BaseSettings):
    """Basic Configuration for definition environment."""

    def __hash__(self):
        return hash((type(self),) + tuple(self.__dict__.values()))

    name: str
    web_api_endpoint: str
    website_endpoint: str
    s3_region: str
    ssl_verify: bool = Field(True, env="TIDY3D_SSL_VERIFY")

    def active(self) -> None:
        """Activate the environment instance."""
        Env.set_current(self)

    def get_real_url(self, path: str) -> str:
        """Get the real url for the environment instance.

        Parameters
        ----------
        path : str
            Base path to append to web api endpoint.

        Returns
        -------
        str
            Full url for the webapi.
        """
        return "/".join([self.web_api_endpoint, path])


dev = EnvironmentConfig(
    name="dev",
    s3_region="us-east-1",
    web_api_endpoint="https://tidy3d-api.dev-simulation.cloud",
    website_endpoint="https://tidy3d.dev-simulation.cloud",
)

uat = EnvironmentConfig(
    name="uat",
    s3_region="us-gov-west-1",
    web_api_endpoint="https://uat-tidy3d-api.simulation.cloud",
    website_endpoint="https://uat-tidy3d.simulation.cloud",
)

uat2 = EnvironmentConfig(
    name="uat2",
    s3_region="us-west-2",
    web_api_endpoint="https://tidy3d-api.uat-simulation.cloud",
    website_endpoint="https://tidy3d.uat-simulation.cloud",
)

prod = EnvironmentConfig(
    name="prod",
    s3_region="us-gov-west-1",
    web_api_endpoint="https://tidy3d-api.simulation.cloud",
    website_endpoint="https://tidy3d.simulation.cloud",
)


class Environment:
    """Environment decorator for user interactive.

    Example
    -------
    >>> Env.dev.active()
    >>> Env.current.name == "dev"
    """

    env_map = dict(
        dev=dev,
        uat=uat,
        uat2=uat2,
        prod=prod,
    )

    def __init__(self):
        """Initialize the environment."""
        env_key = os.environ.get("TIDY3D_ENV")
        env_key = env_key.lower() if env_key else env_key
        log.info(f"env_key is {env_key}")
        if not env_key:
            self._current = prod
        elif env_key in self.env_map:
            self._current = self.env_map[env_key]
        else:
            log.warning(
                f"The value '{env_key}' for the environment variable TIDY3D_ENV is not supported. "
                f"Using prod as default."
            )
            self._current = prod

    @property
    def current(self) -> EnvironmentConfig:
        """Get the current environment.

        Returns
        -------
        EnvironmentConfig
            The config for the current environment.
        """
        return self._current

    @property
    def dev(self) -> EnvironmentConfig:
        """Get the dev environment.

        Returns
        -------
        EnvironmentConfig
            The config for the dev environment.
        """
        return dev

    @property
    def uat(self) -> EnvironmentConfig:
        """Get the uat environment.

        Returns
        -------
        EnvironmentConfig
            The config for the uat environment.
        """
        return uat

    @property
    def uat2(self) -> EnvironmentConfig:
        """Get the uat2 environment.

        Returns
        -------
        EnvironmentConfig
            The config for the uat environment.
        """
        return uat2

    @property
    def prod(self) -> EnvironmentConfig:
        """Get the prod environment.

        Returns
        -------
        EnvironmentConfig
            The config for the prod environment.
        """
        return prod

    def set_current(self, config: EnvironmentConfig) -> None:
        """Set the current environment.

        Parameters
        ----------
        config : EnvironmentConfig
            The environment to set to current.
        """
        self._current = config


Env = Environment()



import numpy as np
import gdstk

import tidy3d as td

from tidy3d.plugins.polyslab import ComplexPolySlab
from ..utils import assert_log_level, log_capture


def test_divide_simple_events():
    """PolySlab division for simple neighbor vertex-vertex type edge events."""

    # simple edge events occurs during erosion
    vertices_ero = ((0, 0), (1, 0), (1, 1), (0, 1), (0, 0.9), (0, 0.11))
    # simple edge events occurs during dilation
    vertices_dil = ((0, 0), (3, 0), (3, 1), (0, 1), (0, 0.9), (0.5, 0.55), (0.5, 0.45), (0, 0.1))

    vertices_list = [vertices_ero, vertices_dil]
    for vertices in vertices_list:
        for angle in [0, np.pi / 4, -np.pi / 4]:
            for reference_plane in ["top", "middle", "bottom"]:
                s = ComplexPolySlab(
                    vertices=vertices,
                    slab_bounds=(0, 1),
                    axis=2,
                    sidewall_angle=angle,
                    reference_plane=reference_plane,
                )
                _ = s.sub_polyslabs
                _ = s.geometry_group


def test_many_sub_polyslabs(log_capture):
    """warn when too many subpolyslabs are generated."""

    # generate vertices that can generate at least this number
    # of sub-polyslabs
    num_subpoly = 200
    dl_list = np.linspace(0, 0.1, num_subpoly)
    vertices = [(sum(dl_list[: i + 1]), 0) for i in range(num_subpoly)]
    vertices = vertices + [(5, 20)]

    s = ComplexPolySlab(
        vertices=vertices,
        slab_bounds=(0, 10),
        axis=2,
        sidewall_angle=np.pi / 4,
        reference_plane="bottom",
    )
    _ = td.Structure(
        geometry=s.geometry_group,
        medium=td.Medium(permittivity=2),
    )
    assert_log_level(log_capture, "WARNING")


def test_divide_simulation():
    """Test adding to a simulation."""

    vertices = ((0, 0), (1, 0), (1, 1), (0, 1), (0, 0.9), (0, 0.11))

    s = ComplexPolySlab(
        vertices=vertices,
        slab_bounds=(0, 1),
        axis=2,
        sidewall_angle=np.pi / 4,
        reference_plane="bottom",
    )
    struct = td.Structure(
        geometry=s.geometry_group,
        medium=td.Medium(permittivity=2),
    )
    _ = td.Simulation(
        run_time=1e-12,
        size=(1, 1, 1),
        grid_spec=td.GridSpec.auto(wavelength=1.0),
        structures=(struct,),
    )
    _ = td.Simulation(
        run_time=1e-12,
        size=(1, 1, 1),
        grid_spec=td.GridSpec.auto(wavelength=1.0),
        structures=(s.to_structure(td.Medium(permittivity=2)),),
    )


def test_gds_import(tmp_path):
    """construct complex polyslabs from gds (mostly from GDSII notebook)"""

    # Waveguide width
    wg_width = 0.45
    # Waveguide separation in the beginning/end
    wg_spacing_in = 8
    # Length of the coupling region
    coup_length = 10
    # Angle of the sidewall: positive values for the base larger than the top
    sidewall_angle = np.pi / 4
    # Reference plane where the cross section of the device is defined
    reference_plane = "bottom"
    # Length of the bend region
    bend_length = 16
    # Waveguide separation in the coupling region
    wg_spacing_coup = 0.10
    # Total device length along propagation direction
    device_length = 100

    def tanh_interp(max_arg):
        """Interpolator for tanh with adjustable extension"""
        scale = 1 / np.tanh(max_arg)
        return lambda u: 0.5 * (1 + scale * np.tanh(max_arg * (u * 2 - 1)))

    def make_coupler(
        length, wg_spacing_in, wg_width, wg_spacing_coup, coup_length, bend_length, npts_bend=30
    ):
        """Make an integrated coupler using the gdstk RobustPath object."""
        # bend interpolator
        interp = tanh_interp(3)
        delta = wg_width + wg_spacing_coup - wg_spacing_in
        offset = lambda u: wg_spacing_in + interp(u) * delta

        coup = gdstk.RobustPath(
            (-0.5 * length, 0),
            (wg_width, wg_width),
            wg_spacing_in,
            simple_path=True,
            layer=1,
            datatype=[0, 1],
        )
        coup.segment((-0.5 * coup_length - bend_length, 0))
        coup.segment(
            (-0.5 * coup_length, 0), offset=[lambda u: -0.5 * offset(u), lambda u: 0.5 * offset(u)]
        )
        coup.segment((0.5 * coup_length, 0))
        coup.segment(
            (0.5 * coup_length + bend_length, 0),
            offset=[lambda u: -0.5 * offset(1 - u), lambda u: 0.5 * offset(1 - u)],
        )
        coup.segment((0.5 * length, 0))
        return coup

    # Create a gds cell to add our structures to
    coup_cell = gdstk.Cell("Coupler")

    # make substrate and add to cell
    substrate = gdstk.rectangle(
        (-device_length / 2, -wg_spacing_in / 2 - 10),
        (device_length / 2, wg_spacing_in / 2 + 10),
        layer=0,
    )

    coup_cell.add(substrate)

    # make coupler and add to the cell
    coup = make_coupler(
        device_length, wg_spacing_in, wg_width, wg_spacing_coup, coup_length, bend_length
    )

    coup_cell.add(coup)

    # Create a library for the cell and save it, just so that we can demosntrate loading
    # geometry from a gds file
    gds_path = tmp_path / "coupler.gds"

    lib = gdstk.Library()
    lib.add(coup_cell)
    lib.write_gds(gds_path)

    lib_loaded = gdstk.read_gds(gds_path)
    coup_cell_loaded = lib_loaded.top_level()[0]

    # Define waveguide height
    wg_height = 0.3
    dilation = 0.02

    [substrate_geo] = ComplexPolySlab.from_gds(
        coup_cell_loaded,
        gds_layer=0,
        gds_dtype=0,
        axis=2,
        slab_bounds=(-430, 0),
        reference_plane=reference_plane,
    )
    _ = ComplexPolySlab.from_gds(
        coup_cell_loaded,
        gds_layer=1,
        axis=2,
        slab_bounds=(0, wg_height),
        sidewall_angle=sidewall_angle,
        dilation=dilation,
    )



import numpy as np

from tidy3d.plugins.fitter import StableDispersionFitter, DispersionFitter


def test_dispersion_load_list():
    """performs a fit on some random data"""
    num_data = 10
    n_data = np.random.random(num_data)
    wvls = np.linspace(1, 2, num_data)
    _ = StableDispersionFitter(wvl_um=wvls, n_data=n_data)


def test_dispersion_load_file():
    """loads dispersion model from nk data file"""
    _ = StableDispersionFitter.from_file("tests/data/nk_data.csv", skiprows=1, delimiter=",")


def test_dispersion_load_url():
    """performs a fit on some random data"""

    # both n and k
    url_csv = "https://refractiveindex.info/data_csv.php?datafile=data/main/Ag/Johnson.yml"
    url_txt = "https://refractiveindex.info/data_txt.php?datafile=data/main/Ag/Johnson.yml"
    fitter = DispersionFitter.from_url(url_csv, delimiter=",")
    fitter = StableDispersionFitter.from_url(url_csv, delimiter=",")
    fitter_txt = DispersionFitter.from_url(url_txt, delimiter="\t")
    fitter_txt = StableDispersionFitter.from_url(url_txt, delimiter="\t")
    fitter_txt.wvl_range = [0.3, 0.8]
    assert len(fitter_txt.freqs) < len(fitter.freqs)

    # only k
    url_csv = "https://refractiveindex.info/data_csv.php?datafile=data/main/N2/Peck-0C.yml"
    url_txt = "https://refractiveindex.info/data_txt.php?datafile=data/main/N2/Peck-0C.yml"
    fitter = DispersionFitter.from_url(url_csv, delimiter=",")
    fitter_txt = DispersionFitter.from_url(url_txt, delimiter="\t")



""" tests converted webapi """
import os
from unittest import TestCase, mock

import tidy3d.web as web
from tidy3d.web.auth import get_credentials, encode_password
from ..utils import SIM_FULL as sim_original

CALLBACK_URL = "https://callbackurl"

""" core webapi """

task_id_global = []

# from tidy3d.web.config import DEFAULT_CONFIG
# DEFAULT_CONFIG.solver_version = "python_upgrade-0.0.0"


def _get_gloabl_task_id():
    """returns the task id from the list"""
    return task_id_global[0]


class Test(TestCase):
    """Use unittest.mock to check that authentication with environment variables works."""

    @mock.patch("tidy3d.web.auth.set_authentication_config")
    def test_get_email_passwd_auth_key_ok(self, set_authentication_config):
        os.environ["TIDY3D_USER"] = "mytestuser"
        os.environ["TIDY3D_PASS"] = "mytestpass"
        get_credentials()
        set_authentication_config.assert_called_with("mytestuser", encode_password("mytestpass"))


def test_webapi_0_run(tmp_path):
    """test complete run"""
    _ = web.run(
        simulation=sim_original, task_name="test_webapi", path=str(tmp_path / "sim_data.hdf5")
    )


def test_webapi_1_upload():
    """test that task uploads ok"""

    task_id = web.upload(
        simulation=sim_original, task_name="test_webapi", callback_url=CALLBACK_URL
    )
    task_info = web.get_info(task_id)
    assert task_info.callbackUrl == CALLBACK_URL
    task_id_global.append(task_id)


def test_webapi_2_get_info():
    """test that we can grab information about task"""
    task_id = _get_gloabl_task_id()
    _ = web.get_info(task_id)


def test_webapi_3_start():
    """test that we can start running task"""
    task_id = _get_gloabl_task_id()
    web.start(task_id)


def test_webapi_4_monitor():
    """test that we can monitor task"""
    task_id = _get_gloabl_task_id()
    web.monitor(task_id)


def test_webapi_5_download(tmp_path):
    """download the simulation data"""
    task_id = _get_gloabl_task_id()
    web.download(task_id, path=str(tmp_path / "sim_data.hdf5"))


def test_webapi_6_load(tmp_path):
    """load the results into sim_data"""
    task_id = _get_gloabl_task_id()
    sim_data = web.load(task_id, path=str(tmp_path / "sim_data.hdf5"))
    first_monitor_name = sim_original.monitors[0].name
    _ = sim_data[first_monitor_name]


def test_webapi_7_download_json():
    """download the simulation data"""
    task_id = _get_gloabl_task_id()
    web.download_json(task_id)


def test_webapi_8_download_log():
    """download the simulation data"""
    task_id = _get_gloabl_task_id()
    web.download_log(task_id)


def test_webapi_9_load_simulation():
    """download the simulation data"""
    task_id = _get_gloabl_task_id()
    sim = web.load_simulation(task_id)
    assert sim == sim_original


# For some reason this test interferes with tests 5 and 6
# def test_webapi_7_delete():
#     """test that we can monitor task"""
#     task_id = _get_gloabl_task_id()
#     web.delete(task_id)
#     with pytest.raises(HTTPError):
#         task_info = web.get_info(task_id)


def test_webapi_8_get_tasks():
    """test that we can get tasks orderd chronologically"""
    tasks = web.get_tasks(num_tasks=5)
    for i in range(4):
        assert tasks[i]["submit_time"] > tasks[i + 1]["submit_time"]


""" Jobs """

jobs_global = []


def _get_gloabl_job():
    """returns the task id from the list"""
    return jobs_global[0]


def test_job_0_run(tmp_path):
    """test complete run"""
    job = web.Job(simulation=sim_original, task_name="test_job", callback_url=CALLBACK_URL)
    job.run(path=str(tmp_path / "sim_data.hdf5"))


def test_job_1_upload():
    """test that task uploads ok"""
    job = web.Job(simulation=sim_original, task_name="test_job")
    jobs_global.append(job)


def test_job_2_get_info():
    """test that we can grab information about task"""
    job = _get_gloabl_job()
    _ = job.get_info()


def test_job_3_start():
    """test that we can start running task"""
    job = _get_gloabl_job()
    job.start()


def test_job_4_monitor():
    """test that we can monitor task"""
    job = _get_gloabl_job()
    job.monitor()


def test_job_5_download(tmp_path):
    """download the simulation data"""
    job = _get_gloabl_job()
    job.download(path=str(tmp_path / "sim_data.hdf5"))


def test_job_6_load(tmp_path):
    """load the results into sim_data"""
    job = _get_gloabl_job()
    sim_data = job.load(path=str(tmp_path / "sim_data.hdf5"))
    first_monitor_name = sim_original.monitors[0].name
    _ = sim_data[first_monitor_name]


def _test_job_7_delete():
    """test that we can monitor task"""
    job = _get_gloabl_job()
    job.delete()
    task_info = job.get_info()
    assert task_info.status in ("deleted", "deleting")


""" Batches """

batches_global = []


def _get_gloabl_batch():
    """returns the task id from the list"""
    return batches_global[0]


def test_batch_0_run(tmp_path):
    """test complete run"""
    sims = 2 * [sim_original]
    simulations = {f"task_{i}": sims[i] for i in range(len(sims))}
    batch = web.Batch(simulations=simulations)
    batch.run(path_dir=str(tmp_path))


def test_batch_1_upload():
    """test that task uploads ok"""
    sims = 2 * [sim_original]
    simulations = {f"task_{i}": sims[i] for i in range(len(sims))}
    batch = web.Batch(simulations=simulations)
    batches_global.append(batch)


def test_batch_2_get_info():
    """test that we can grab information about task"""
    batch = _get_gloabl_batch()
    _ = batch.get_info()


def test_batch_3_start():
    """test that we can start running task"""
    batch = _get_gloabl_batch()
    batch.start()


def test_batch_4_monitor():
    """test that we can monitor task"""
    batch = _get_gloabl_batch()
    batch.monitor()


def test_batch_5_download(tmp_path):
    """download the simulation data"""
    batch = _get_gloabl_batch()
    batch.download(path_dir=str(tmp_path))


def test_batch_6_load(tmp_path):
    """load the results into sim_data"""
    batch = _get_gloabl_batch()
    sim_data_dict = batch.load(path_dir=str(tmp_path))
    first_monitor_name = sim_original.monitors[0].name
    for _, sim_data in sim_data_dict.items():
        _ = sim_data[first_monitor_name]


def test_batchdata_7_load(tmp_path):
    """load the BatchData from file."""
    _ = web.BatchData.load(path_dir=str(tmp_path))


def _test_batch_8_delete():
    """test that we can monitor task"""
    batch = _get_gloabl_batch()
    batch.delete()
    for job in batch:
        task_info = job.get_info()
        assert task_info.status in ("deleted", "deleting")


""" Async """


def test_web_run_async():
    """test complete run"""
    sims = 2 * [sim_original]
    batch_data = web.run_async(sims)
    first_monitor_name = sim_original.monitors[0].name
    for _, sim_data in batch_data.items():
        _ = sim_data[first_monitor_name]
